=begin
###
Copyright 2011-2012 TIG (c)
Permission to use, copy, modify, and distribute this software for any 
purpose, and currently without fee, is hereby granted, provided that  
this text and the above copyright (c) notice appear in all copies.
This software is provided "as is" and without any express or implied 
warranties, including, without limitation, the implied warranties of 
merchantability and fitness for a particular purpose.
###
Parts are based on ThomThom's 'solid_inspector' tool which can also be 
used prior to this tool, so that you get a better understanding of what 
is wrong with your Non-Solid Group...
###
Name: TIG-solidsolver.rb
###
Menu: 'Tools' > 'SolidSolver'
Context-menu  > 'SolidSolver'
###
Usage: [Note this works on >= v8 only]
Select a Non-Solid Group/Component-Instance and run the tool from the menu
or context-menu.
If it contains nested groups or instances you are warned, and given the 
option to explode these - it could never be a 'solid' otherwise.
It analyzes the object's form for holes [which it heals], small faces 
[also healed] and removes partitions/flaps that are making the form 
'non-manifold', and it also removes 'unneeded' unfaced/coplanar edges etc
and finally it merges any 'split' colinear edges.
It then reports on the results 'solidity' & what was done to achieve it.
It is one step undoable.
If the object remains 'non-solid' it's likely be due to some edges still 
having more than two faces - e.g. two otherwise solid cubes touching at 
an edge will therefore use that edge for four faces, which is therefore 
regarded as making it a non-solid.  An additional dialog explains, and it 
then offers the chance to try to rectify this. A 'fix' might produce odd 
results - for example the common edge to two cubes might be removed so 
there are no longer any edges with more than two faces, but the form 
might then be folded back on itself to achieve interpenetrating faces - 
which, although then regarded as forming a 'solid', is an 'impossible 
object' in normal three dimensions... You are given a separate undo for 
this step if you do try it.
Tip: IF you have this type of form temporarily group one of the 'cubes' 
and move if over say 0.5mm, so that there are then two distinct edges 
and it will become a proper solid of two separated volumes, without the 
involuted faces issue.
###
Donations: PayPal.com to info @ revitrev.org
###
Version:
1.0 20111013 First issue.
1.1 20111014 Nested groups now considered.
1.2 20111016 Unsolvable groups etc now spotted.
1.3 20111019 Split colinear edges now merged.
1.4 20111027 Typo in small-hole-heal code fixed.
1.5 20121012 Works on groups/component=instances, makes the result unique.
1.6 20121024 Improved internal partition removal even when they are not full 
			 width. Context-menu now available if appropriate selection.
			 Progress ... is now indicated in the status-bar.
###
=end
###
require 'sketchup.rb'
###
module TIG
  ###
  unless file_loaded?(__FILE__)
    cmd=UI::Command.new("SolidSolver"){self.solidsolver()}
    UI.menu("Tools").add_item(cmd)
	UI.add_context_menu_handler{|menu|
		if self.solidsolver_unsolid?()
			menu.add_item(cmd)
		end
	}	
  end
  file_loaded(__FILE__)
  ###
  def self.solidsolver_unsolid?()
	@model=Sketchup.active_model
	@view=@model.active_view
	@ss=@model.selection
	if @ss.empty? or @ss[1] or (@ss[0].class!=Sketchup::Group and @ss[0].class!=Sketchup::ComponentInstance)
		return nil
    else
		if @ss[0].manifold?
			return false
		else
			return true
		end
	end
  end
  ###
  def self.solidsolver()
    ###
    @model=Sketchup.active_model
	@ss=@model.selection
    if self.solidsolver_unsolid?()==nil
      UI.messagebox("Select ONE GROUP/INSTANCE to make 'Solid'.")
      return nil
    else
      @container=@ss[0]
      if self.solidsolver_unsolid?()==false
        UI.messagebox("The Selected GROUP/INSTANCE is already 'Solid' !")
        return nil
      end
	  @model.start_operation("SolidSolver")
	  @ss.clear
	  @view.refresh
	  @msgX="SolidSolver: .";Sketchup::set_status_text(@msgX)
	  if @container.class==Sketchup::Group
		@entities=@container.entities
		name=@container.name
		uniq="Group is "
		if @entities.parent.instances[1]
			@container.make_unique
			@entities=@container.entities
			if name.empty?
				uniq="Group was made_unique,\nit is "
			else
				uniq="Group \""+name+"\" was made_unique as \""+@container.name+"\",\nit is "
			end
		end
	  else ### component instance
	    @entities=@container.definition.entities
		name=@container.definition.name
		uniq="Component-Instance is "
		if @entities.parent.instances[1]
			@container.make_unique
			@entities=@container.definition.entities
			uniq="Component-instance \""+name+"\" was made_unique as \""+@container.definition.name+"\"\nit is "
		end
	  end
      @ss.clear
      @msg="SolidSolver:\n\nThe Selected "
	  @msg<<uniq
      @cops=0
      @hole=0
      @small=0
      @parts=0
      @splits=0
      self.solidsolver_nested()
	  @msgX<<".";Sketchup::set_status_text(@msgX)
	  @view.refresh
	  self.solidsolver_heal_holes()
	  @msgX<<".";Sketchup::set_status_text(@msgX)
	  @view.refresh
      self.solidsolver_heal_small()
	  @msgX<<".";Sketchup::set_status_text(@msgX)
	  @view.refresh
      self.solidsolver_remove_partitions()
	  @msgX<<".";Sketchup::set_status_text(@msgX)
	  @view.refresh
      self.solidsolver_heal_splits()
	  @msgX<<".";Sketchup::set_status_text(@msgX)
	  @view.refresh
      ###
      if @cops==0 and @hole==0 and @small==0 and @parts==0
        now=""
        still=""
      else
        now="now "
        still="still "
      end
      if @container.manifold?
        @msg<<now+"'Solid'.\n\n"
        isnot=false
      else
        @msg<<still+"NOT 'Solid' !\n\n"
        isnot=true
      end
      ###
      ces="s"
      hes="s"
      ses="s"
      pes="s"
      ces="" if @cops==1
      hes="" if @hole==1
      ses="" if @small==1
      pes="" if @parts==1
      @splits=0 if @splits==1
      @msg<<@gps
      @msg<<@cops.to_s+"\t Unfaced/Coplanar Edge"+ces+" Removed\n"
      @msg<<@hole.to_s+"\t Hole"+hes+" Healed\n"
      @msg<<@small.to_s+"\t Small Face"+ses+" Healed\n"
      @msg<<@parts.to_s+"\t Partition/Flap"+pes+" Removed\n"
      @msg<<@splits.to_s+"\t Split-Edges Merged\n"
      UI.messagebox(@msg)
      if isnot
        edges1=[]
        edges2=[]
        @entities.each{|e|
          edges1 << e if e.is_a?(Sketchup::Edge) and e.faces.length==1
          edges2 << e if e.is_a?(Sketchup::Edge) and e.faces.length>2
        }
        if edges2[0]
          @model.commit_operation
          UI.beep
          if UI.messagebox("SolidSolver:\n\nThe lack of 'Solidity' might be because\nmore than two Faces share some Edges,\n[e.g. two otherwise 'solid' cubes might share an edge].\n\nDo you want to try and fix it?\n\n[It might cause unexpected results, but it will be undoable...]",MB_YESNO,"")==6 ### 6=YES 7=NO
            @model.start_operation("SolidSolver 2")
              self.solidsolver_erase_shared()
            @model.commit_operation
          end
        elsif not @entities[0]
          UI.messagebox("SolidSolver:\n\nThis Group/Instance can never be solved into a 'Solid'!\nTry manually adding/removing Edges/Faces...\n[No changes have been made]")
          @model.abort_operation
        elsif edges1[0]
          UI.messagebox("SolidSolver:\n\nThis Group/Instance can never be solved intp a 'Solid'!\nTry manually adding/removing Edges/Faces...\n[Undo if appropriate]")
          @model.commit_operation
        else
          ents=[]
          @entities.each{|e|ents << e.class}
          if ents.include?(Sketchup::Group) or ents.include?(Sketchup::ComponentInstance)
            UI.messagebox("SolidSolver:\n\nThis Group/Instance still contains Groups/Instances!\nIt can never be considered 'Solid'...")
          end
          @model.commit_operation
        end
      end
	  @ss.add(@container)
    end
  end
    ###
    def self.solidsolver_heal_splits()
      aedges=[]
      @entities.each{|e|aedges << e if e.valid? and e.is_a?(Sketchup::Edge)}
      verts=[]
      aedges.each{|edge|verts << edge.vertices}
      verts.flatten!
      verts.uniq!
      vends=[]
      verts.each{|vx|
        edges=vx.edges
        next unless edges.length==2
		e0=edges[0]
		e2=edges[1]
        ptx=vx.position
		v0=e0.other_vertex(vx)
		v2=e2.other_vertex(vx)
		pt0=v0.position
		pt2=v2.position
        next unless ptx.on_line?([pt0,pt2])
        e0=vx.edges[0]
		vec=pt0.vector_to(pt2)
        axes=nil
		axes=vec.axes if vec.valid?
        vends << [vx.to_s, ptx.to_a, axes]
      }
      vends.to_a.each{|ar|
        vx,pt,axes=ar
	    next unless axes
	    pte=Geom::Point3d.new(pt.x,pt.y,pt.z).offset(axes[0],0.01)
	    edg=@entities.add_line(pt,pte)
	    edg.erase!
      }
      @splits=vends.length+1
    end
    def self.solidsolver_nested()
      @gps=""
      gps=[]
      @entities.each{|e|gps << e if e.class==Sketchup::Group or e.class==Sketchup::ComponentInstance}
      if gps.length==1
        UI.beep
        if UI.messagebox("SolidSolver:\n\nThis Group/Instance contains 1 nested Group/Instance.\n\nDo you want to Explode it ?\n\n[This Group/Instance with never be 'Solid' otherwise...]",MB_YESNO,"")==6 ### 6=YES 7=NO
          num=0
          gone=false
          while not gone
            @entities.to_a.each{|e|
              if e.class==Sketchup::Group or e.class==Sketchup::ComponentInstance
                e.explode
                num+=1
              end
            }
            ents=[]; @entities.each{|e|ents << e.class}; ents.uniq!
            gone=true if not ents.include?(Sketchup::Group) and not ents.include?(Sketchup::ComponentInstance)
          end
          if num==1
            @gps=num.to_s+"\t Nested Group/Instance Exploded\n"
          else
            @gps=num.to_s+"\t Nested Groups/Instances Exploded\n"
          end
        end
      elsif gps.length>1
        if UI.messagebox("SolidSolver:\n\nThis Group/Instance contains "+gps.length.to_s+" nested Groups/Instances.\n\nDo you want to Explode them ?\n\n[This Group/Instance with never be 'Solid' otherwise...]",MB_YESNO,"")==6 ### 6=YES 7=NO
          num=0
          gone=false
          while not gone
            @entities.to_a.each{|e|
              if e.class==Sketchup::Group or e.class==Sketchup::ComponentInstance
                e.explode
                num+=1
              end
            }
            ents=[]; @entities.each{|e|ents << e.class}; ents.uniq!
            gone=true if not ents.include?(Sketchup::Group) and not ents.include?(Sketchup::ComponentInstance)
          end
          @gps=num.to_s+"\t Nested Groups/Instances Exploded\n"
        end
      end
    end
    def self.solidsolver_erase_shared()
	  @msgX<<" .";Sketchup::set_status_text(@msgX)
      @shared=0
      @entities.to_a.each{|e|
        if e.is_a?(Sketchup::Edge) and e.faces.length>2
          e.erase!
          @shared+=1
        end
      }
	  @msgX<<" .";Sketchup::set_status_text(@msgX)
      @entities.to_a.each{|e|
        if e.is_a?(Sketchup::Edge) and e.faces.length==1
          e.find_faces
        end
      }
	  @msgX<<" .";Sketchup::set_status_text(@msgX)
      @msg2="SolidSolver 2:\n\n"
      if @container.manifold?
        @msg2<<"This Group/Instance is now 'Solid'.\n\n"
        isnot=false
      else
        @msg2<<"This Group/Instance is still NOT 'Solid' !\n\n"
        isnot=true
      end
      ees="s"
      ees="" if @shared==1
      UI.messagebox(@msg2+"Removed "+@shared.to_s+" Edge"+ees+" with more than two Faces .\n\n[Use 'Undo' to Revert]")
    end
	def self.solidsolver_erase_coplanar()
	  @msgX<<" .";Sketchup::set_status_text(@msgX)
	  @entities.to_a.each{|e|
        next unless e.valid? and e.is_a?(Sketchup::Edge) 
        if not e.faces[0]
          e.erase!
          @cops+=1
        elsif e.faces.length==2 and e.faces[0].normal.dot(e.faces[1].normal) > 0.99999999 and e.faces[0].material==e.faces[1].material and e.faces[0].back_material==e.faces[1].back_material
          e.erase!
          @cops+=1
        end
      }
	end
    def self.solidsolver_heal_holes()
      edges=[]
      @entities.to_a.each{|e|
        next unless e.valid? and e.is_a?(Sketchup::Edge)
        if e.faces.length==1
          edges << e
        end
      }
      edges.each{|e|
        if e.faces.length==1
          e.find_faces
          @hole+=1
        end
      }
      self.solidsolver_erase_coplanar()
    end
	def self.solidsolver_heal_small()
	    @entities.to_a.each{|e|
          next unless e.valid? and e.is_a?(Sketchup::Edge)
          if e.length <=0.5.mm
            vs=e.start
            ve=e.end
            ps=vs.position
            pe=ve.position
            vec=pe.vector_to(ps)
            @entities.transform_by_vectors(ve, vec)
            pp=vs.position
            tgp=@entities.add_group()
            li=tgp.entities.add_line(pp, pp.offset(Z_AXIS))
            tgp.entities.transform_by_vectors([li.end], [Z_AXIS.reverse])
            tgp.explode
            @small+=1
          else
            e.find_faces
            e.erase! if e.valid? and e.faces.length==2 and e.faces[0].normal.dot(e.faces[1].normal) > 0.99999999 and e.faces[0].material==e.faces[1].material and e.faces[0].back_material==e.faces[1].back_material
          end
        }
		self.solidsolver_erase_coplanar()
	end
    def self.solidsolver_remove_partitions()
		stack=@entities.select{|e|e.is_a?(Sketchup::Edge) && e.faces.length!=2}
		while stack[0]
			cluster=[]
			cluster << stack.shift
			edge = cluster.first
			haystack = ([edge.start.edges + edge.end.edges]-[edge]).first & stack
			until haystack.empty?
				e=haystack.shift
				if stack.include?(e)
					cluster << e
					stack.delete(e)
					haystack+=([e.start.edges + e.end.edges]-[e]).first & stack 
				end
			end
		    ###
		    if cluster[0]
				facs=[]
				cluster.each{|e|
					next unless e.valid?
					facs << e.faces
				}
				facs.flatten!
				facs.uniq!
				sus=[]
				facs.each{|f|
					crap=true
					f.edges.each{|e|
						if not cluster.include?(e)
							crap=false
							break
						end
					}
					next unless crap
					sus << f
				}
				if sus[0]
				  sus.uniq!
				  @entities.erase_entities(sus)
				  @parts=@parts+sus.length
				end
				###
				self.solidsolver_erase_coplanar()
				###
		    end#if
		    ###
		end#stack
		###
		self.solidsolver_erase_coplanar()
		@msgX<<" .";Sketchup::set_status_text(@msgX)
		### repeat for odds
		tc=@container.transformation
		e3s=@entities.select{|e|e.is_a?(Sketchup::Edge) && e.faces.length>2}
		if e3s[0] ### don't bother if no partitions !
			togos=[]
			e3s.each{|e|
				@msgX<<" .";Sketchup::set_status_text(@msgX)
				e.faces.each{|f|
					pt=f.bounds.center.transform(tc)
					ray=@model.raytest(pt, f.normal.transform(tc))
					next if not ray or not ray[1].include?(@container)
					fa=ray[1][-1]
					fa=fa.faces[0] unless fa.is_a?(Sketchup::Face)
					togos << f if f.normal.transform(tc).angle_between(fa.normal.transform(tc))<90.degrees
				}
			}
			@msgX<<" .";Sketchup::set_status_text(@msgX)
			if togos[0]
				togos.uniq!
				@entities.erase_entities(togos)
				@parts=@parts+togos.length
			end
			self.solidsolver_erase_coplanar()
		end
		###
    end
    ###
end#module
